@@ -1021,6 +1021,8 @@ _mx_widget_propagate_disabled (ClutterContainer *container,
       if (CLUTTER_IS_CONTAINER (child))
         _mx_widget_propagate_disabled ((ClutterContainer *) child, disabled);
     }
+
+  g_list_free (children);
 }
 
 static void
@@ -1058,6 +1060,18 @@ mx_widget_parent_set (ClutterActor *actor,
     }
 }
 
+static gboolean
+mx_widget_get_paint_volume (ClutterActor       *actor,
+                            ClutterPaintVolume *volume)
+{
+  /* the allocation of scrollable widgets cannot be used as the paint volume
+   * because it does not account for any transformations applied during
+   * scrolling */
+  if (MX_IS_SCROLLABLE (actor))
+    return FALSE;
+
+  return clutter_paint_volume_set_from_allocation (volume, actor);
+}
 
 static void
 mx_widget_class_init (MxWidgetClass *klass)
@@ -1087,6 +1101,8 @@ mx_widget_class_init (MxWidgetClass *klass)
   actor_class->hide = mx_widget_hide;
   actor_class->parent_set = mx_widget_parent_set;
 
+  actor_class->get_paint_volume = mx_widget_get_paint_volume;
+
   klass->paint_background = mx_widget_real_paint_background;
 
   /* stylable interface properties */
@@ -1770,7 +1786,7 @@ mx_widget_set_tooltip_delay (MxWidget *widget,
 guint
 mx_widget_get_tooltip_delay (MxWidget *widget)
 {
-  g_return_if_fail (MX_IS_WIDGET (widget));
+  g_return_val_if_fail (MX_IS_WIDGET (widget), 0);
 
   return widget->priv->tooltip_delay;
 }
